# $Id: Job.pm,v 1.5 2002/02/08 16:04:58 joern Exp $

#-----------------------------------------------------------------------
# Copyright (C) 2001-2002 Jrn Reder <joern@zyn.de> All Rights Reserved
# 
# This program is part of Video::DVDRip, which is free software; you can
# redistribute it and/or modify it under the same terms as Perl itself.
#-----------------------------------------------------------------------

package Video::DVDRip::Cluster::Job;

use base Video::DVDRip::Base;

use Carp;
use strict;

sub id				{ shift->{id}				}

sub node			{ shift->{node}				}
sub project			{ shift->{project}			}

sub event_waiter		{ shift->{event_waiter}			}
sub set_event_waiter		{ shift->{event_waiter}		= $_[1]	}

sub chunk_nr			{ shift->{chunk_nr}			}
sub set_chunk_nr		{ shift->{chunk_nr}		= $_[1]	}

sub stopping_in_progress	{ shift->{stopping_in_progress}		}
sub set_stopping_in_progress	{ shift->{stopping_in_progress}	= $_[1] }

# internal job ID
my $JOB_ID = 0;

# hash of running jobs (needed for proper cleanup management)
my %RUNNING_JOBS;

# projects want's to know on which nodes they're on
my %NODES_PER_PROJECT;

sub new {
	my $class = shift;
	my %par = @_;
	my  ($project, $node) =
	@par{'project','node'};
	
	# build object
	my $self = bless {
		id           => ++$JOB_ID,
		project      => $project,
		node         => $node,
	}, $class;

	# register that this node is working for this project
	$NODES_PER_PROJECT{$project->id}->{$node->name} = $node;

	# register this job in the running jobs hash
	$RUNNING_JOBS{$self->id} = $self;

	# assign this job to the node
	$node->set_assigned_job ($self);
	
	return $self;
}

sub DESTROY {
	shift->remove;
}

sub remove {
	my $self = shift;

	# do nothing if already removed (e.g. when called by
	# the DESTROY handler after we removed us already)
	return 1 if not $RUNNING_JOBS{$self->id};

	# reset node's state
	my $node = $self->node;
	$node->set_state('idle');
	$node->set_assigned_job (undef);
	$node->stop if $self->stopping_in_progress;

	# remove this node from the node per project hash
	my $project = $self->project;
	delete $NODES_PER_PROJECT{$project->id}->{$node->name};

	# reset project state to 'waiting' if no other nodes
	# are currtntly transcoding this project
	$project->set_state ('waiting')
		if $project->state eq 'transcoding' and
		   keys %{$self->get_project_nodes(project => $project)} == 0;

	# delete us from the RUNNING_JOBS hash
	delete $RUNNING_JOBS{$self->id};

	1;
}

sub abort {
	my $self = shift;

	# We call the Event->io callback with 1 as second parameter.
	# This aborts the the waiter session and the corresponding
	# finished callback will be called.
	my $cb = $self->event_waiter->cb;
	&$cb (undef, 1);

	1;
}

sub get_project_nodes {
	my $thing = shift;
	my %par = @_;
	my ($project) = @par{'project'};
	
	return $NODES_PER_PROJECT{$project->id};
}

sub start_transcoding {
	my $self = shift;
	my %par = @_;
	my  ($chunk_nr, $pass, $frames_cnt) =
	@par{'chunk_nr','pass','frames_cnt'};

	$pass       ||= 1;
	$frames_cnt ||= 0;

	$self->set_chunk_nr ( $chunk_nr );

	my $node    = $self->node;
	my $project = $self->project;

	# get transcode command
	$project->set_assigned_job ( $self );
	my $title   = $project->title;
	my $command = $title->get_transcode_command ( pass => $pass );
	$project->set_assigned_job ( undef );

	my $log_multipass      = $title->tc_multipass ? " (multipass $pass)" : "";
	my $mp_progress_factor = $title->tc_multipass ? $pass - 1 : 0;
	my $mp_sum_factor      = $title->tc_multipass ? 2 : 1;

	$self->log (
		"Start transcode job$log_multipass for project '".
		$node->name.
		"' with chunk $chunk_nr on node '".
		$node->name."'"
	);
	
	# the node needs this information for reporting
	# its proper progress state
	$node->set_assigned_chunk ( $chunk_nr );
	$node->set_progress_frames( $frames_cnt ); # frames_cnt is set for 2nd pass

	# get code for command execution on this node
	$command = $node->get_popen_code ( command => $command );

	$self->log (3, "transcode popen command: $command");

	# start a new session for this command
	$node->set_progress_start_time (time) if $pass != 2;
	
	my $master = Video::DVDRip::Cluster::Master->get_master;

	my $frames_cnt;
	my $successful_finished = 0;

	my $waiter = $master->popen (
		command      => $command,
		cb_line_read => sub {
			my ($line) = @_;
			if ( $line =~ /split.*?frames.*?-c\s+\d+-(\d+)/ ) {
				$node->set_progress_frames_cnt (
					$1 * $mp_sum_factor
				);
				$frames_cnt = $1;

			} elsif ( $line =~ /\[\d{6}-(\d+)\]/ ) {
				$node->set_progress_frames(
					$1 +
					$mp_progress_factor * $frames_cnt
				);

			} elsif ( $line =~ /DVDRIP_SUCCESS/ ) {
				$successful_finished = 1;
			}
		},
		cb_finished  => sub {
			if ( $successful_finished ) {
				if ( $title->tc_multipass and $pass == 1 ) {
					$self->start_transcoding (
						chunk_nr   => $chunk_nr,
						pass       => 2,
						frames_cnt => $frames_cnt,
					);
				} else {
					$self->commit_transcoding;
				}
			} else {
				$self->abort_transcoding;
			}
		},
	);

	$self->set_event_waiter ( $waiter );

	# change node and project state
	$node->set_state ( 'transcoding' );
	$project->set_state ( 'transcoding' );
	
	1;
}

sub commit_transcoding {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	my $mp_factor = $title->tc_multipass ? 2 : 1;

	$self->log (
		"Successfully finished transcode job ".
		"for project '".$project->label.
		"' with chunk $chunk_nr on node '".
		$node->name."'"
	);

	$title->set_frames_finished (
		$title->frames_finished +
		$title->frames_per_chunk * $mp_factor
	);

	$title->set_chunk_state (
		nr    => $chunk_nr,
		state => 'finished'
	);

	$self->move_chunk_file;

	$self->remove;

	Video::DVDRip::Cluster::Master->get_master->job_control;

	1;
}

sub move_chunk_file {
	my $self = shift;
	
	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	# get code for command execution on this node
	$project->set_assigned_job ( $self );
	my $command = $title->get_move_avi_command;
	$command = $node->get_popen_code ( command => $command );
	$project->set_assigned_job ( undef );
	
	$self->log ("Moving avi file to its final destination: $command");
	system ($command);
	
	1;
}

sub abort_transcoding {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	$self->log (
		"Transcode job aborted ".
		"for project '".$project->label.
		"' with chunk $chunk_nr on node '".
		$node->name."'"
	);

	$title->set_chunk_state (
		nr    => $chunk_nr,
		state => 'todo'
	);

	$self->remove;

	1;
}

sub start_merging {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;

	$self->log (
		"Start merging job for project '".
		$project->label.
		"' on node '".$node->name."'"
	);

	# the node needs this information for reporting
	# its proper progress state
	$node->set_progress_frames( 0 );

	# get merge command
	$project->set_assigned_job ( $self );
	my $title   = $project->title;
	my $command = $title->get_merge_command;
	$project->set_assigned_job ( undef );
	
	# get code for command execution on this node
	$command = $node->get_popen_code ( command => $command );

	$self->log (3, "merge popen command: $command");

	# start a new session for this command
	$node->set_progress_start_time (time);
	
	my $master = Video::DVDRip::Cluster::Master->get_master;

	my $successful_finished = 0;

	my $waiter = $master->popen (
		command      => $command,
		cb_line_read => sub {
			my ($line) = @_;
			if ( $line =~ /done\s+merging\s+\d+\s+files/ ) {
				$successful_finished = 1;
			} elsif ( $line =~ /file\s+(\d+)\s+/ ) {
				$node->set_progress_merge($1);
			}
		},
		cb_finished  => sub {
			if ( $successful_finished ) {
				$self->commit_merging;
			} else {
				$self->abort_merging;
			}
		},
	);

	$self->set_event_waiter ( $waiter );

	# change node and project state
	$node->set_state ( 'merging' );
	$project->set_state ( 'merging' );
	
	1;
}

sub commit_merging {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	
	$self->log (
		"Succesfully finished merge job for project '".
		$project->label."' on node '".$node->name."'"
	);
	
	$project->set_state ("merged");
	$project->save;
	
	$self->remove;

	Video::DVDRip::Cluster::Master->get_master->job_control;

	1;
}

sub abort_merging {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	
	$self->log (
		"Merge job aborted for project '".
		$project->label.
		"' on node '".$node->name."'"
	);
	
	$project->force_state ("transcoded");
	$project->save;
	
	$self->remove;

	1;
}

sub start_audio_processing {
	my $self = shift;

	my $node    = $self->node;
	my $project = $self->project;

	$self->log (
		"Start audio processing job for project '".
		$project->label."' on node '".$node->name."'"
	);

	# the node needs this information for reporting
	# its proper progress state
	$node->set_progress_frames( 0 );

	# get audio processing command
	$project->set_assigned_job ( $self );
	my $title   = $project->title;
	my $command = $title->get_process_audio_command;
	$project->set_assigned_job ( undef );
	
	# get code for command execution on this node
	$command = $node->get_popen_code ( command => $command );

	$self->log (3, "audio processing popen command: $command");

	# start a new session for this command
	$node->set_progress_start_time (time);
	
	my $master = Video::DVDRip::Cluster::Master->get_master;

	my $successful_finished = 0;

	my $waiter = $master->popen (
		command      => $command,
		cb_line_read => sub {
			my ($line) = @_;
			if ( $line =~ /split.*?frames\s+(\d+)-(\d+)/ ) {
				$node->set_progress_frames_cnt ($2-$1);
			} elsif ( $line =~ /encoding.*?\[(\d{6})-(\d+)\]/ ) {
				$node->set_progress_frames($2-$1);
			} elsif ( $line =~ /DVDRIP_SUCCESS/ ) {
				$successful_finished = 1;
			}
		},
		cb_finished  => sub {
			if ( $successful_finished ) {
				$self->commit_audio_processing;
			} else {
				$self->abort_audio_processing;
			}
		},
	);

	$self->set_event_waiter ( $waiter );

	# change node and project state
	$node->set_state ( 'audio processing' );
	$project->set_state ( 'audio processing' );
	
	1;
}

sub commit_audio_processing {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	$self->log (
		"Succesfully finished audio processing job for project '".
		$project->label."' on node '".$node->name."'"
	);

	$project->set_state ("audio processed");
	$project->set_state ("finished")
		if not $project->title->with_avisplit;

	$project->save;

	$self->remove;

	Video::DVDRip::Cluster::Master->get_master->job_control;

	1;
}

sub abort_audio_processing {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	$self->log (
		"Audio processing job aborted ".
		"for project '".$project->label.
		"' on node '".$node->name."'"
	);

	$project->force_state ("merged");
	$project->save;

	$self->remove;

	1;
}

sub start_avisplit {
	my $self = shift;

	my $node    = $self->node;
	my $project = $self->project;

	$self->log (
		"Start avisplit job for project '".
		$project->label."' on node '".$node->name."'"
	);

	# get audio processing command
	$project->set_assigned_job ( $self );
	my $title   = $project->title;
	my $command = $title->get_split_command;
	$project->set_assigned_job ( undef );
	
	# the node needs this information for reporting
	# its proper progress state
	$node->set_progress_frames( 0 );
	$node->set_progress_frames_cnt ( $title->frames );

	# get code for command execution on this node
	$command = $node->get_popen_code ( command => $command );

	$self->log (3, "avisplit popen command: $command");

	# start a new session for this command
	$node->set_progress_start_time (time);
	
	my $master = Video::DVDRip::Cluster::Master->get_master;

	my $successful_finished = 0;

	my $waiter = $master->popen (
		command      => $command,
		cb_line_read => sub {
			my ($line) = @_;
			if ( $line =~ /\(\d{6}-(\d+)\),\s+(.*?)\[.*?$/ ) {
				$node->set_progress_frames_cnt ($1);
			} elsif ( $line =~ /DVDRIP_SUCCESS/ ) {
				$successful_finished = 1;
			}
		},
		cb_finished  => sub {
			if ( $successful_finished ) {
				$self->commit_avisplit;
			} else {
				$self->abort_avisplit;
			}
		},
	);

	$self->set_event_waiter ( $waiter );

	# change node and project state
	$node->set_state ( 'splitting avi' );
	$project->set_state ( 'splitting avi' );
	
	1;
}

sub commit_avisplit {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	$self->log (
		"Succesfully finished avisplit job for project '".
		$project->label."' on node '".$node->name."'"
	);

	$project->set_state ("avi splitted");
	$project->set_state ("finished");

	$project->save;

	$self->remove;

	Video::DVDRip::Cluster::Master->get_master->job_control;

	1;
}

sub abort_avisplit {
	my $self = shift;

	my $node     = $self->node;
	my $project  = $self->project;
	my $title    = $project->title;
	my $chunk_nr = $self->chunk_nr;

	$self->log (
		"avisplit job aborted ".
		"for project '".$project->label.
		"' on node '".$node->name."'"
	);

	$project->force_state ("audio processed");
	$project->save;

	$self->remove;

	1;
}

1;
